package;

import format.png.Data;
import format.png.Tools;
import format.png.Writer;
import haxe.io.Bytes;
import haxe.io.BytesInput;
import haxe.io.BytesOutput;
import haxe.io.Eof;
import haxe.io.Input;
import haxe.io.Path;
import sys.FileSystem;
import sys.io.File;
import sys.io.FileInput;
import sys.io.FileOutput;
import sys.io.FileSeek;

/**
 * NUKA-CEL! by Athena/ユウリ マヤ
 */
class Main
{
	private var path:String;
	private var countART2:Int;
	private var countCO2R:Int;
	private var countCLC2:Int;
	private var countCL3L:Int;
	private var countSDL:Int;
	
	static public function main( )
	{
		new Main( );
	}

	public function new()
	{
		this.countART2 = 0;
		this.countCO2R = 0;
		this.countCLC2 = 0;
		this.countCL3L = 0;
		this.countSDL = 0;

		var args = Sys.args( );
		if ( args.length == 0 ) {
			trace( "usage: nukacel [file]" );
			return;
		}

		var file = args[0];
		this.path = Path.withoutExtension( Path.withoutDirectory( file ) );

		FileSystem.createDirectory( path );
		FileSystem.createDirectory( path+"/ART2" );
		FileSystem.createDirectory( path+"/CO2R" );
		FileSystem.createDirectory( path+"/CLC2" );
		FileSystem.createDirectory( path + "/UNKNOWN" );
		FileSystem.createDirectory( path+"/CL3L" );
		// FileSystem.createDirectory( path+"/SDL" );

		// 開く
		var f:FileInput = File.read( file );
		f.bigEndian = false;

		// ファイルテーブル読み込み
		var fileTable = new Array<{addr:Int, size:Int, fileId:Int}>( );
		var remain:Int = 0xA000;
		var addr:Int = 0xA000;
		while ( 0 < remain ) {
			var v = f.readInt32( );
			var fileId:Int = v & 0x7ff;
			var size:Int = ( ( v >> 12 ) * 128 );
			if ( size == 0 ) break;
			trace( 'file ID: $fileId, src: $v, address: $addr, size: $size' );
			fileTable.push( { addr:addr, size:size, fileId:fileId } );
			addr += size;
			remain -= 4;
		}

		// データを読み込む
		for( t in fileTable ) {
			f.seek( t.addr, FileSeek.SeekBegin );
			var bi = new BytesInput( f.read( t.size ) );
			var header = bi.readInt32( );
			bi.position = 0;

			if ( header == 0x08088B1f ) {
				// 圧縮データ もっとNUGA-CEL!用
				trace( "header: compressed data" );
				this.extractCompressedData( bi, t.fileId );
			}else {
				// 無圧縮
				this.extractData( bi, t.fileId );
			}
		}

		f.close( );

		trace( "done!" );
	}

	/**
	 * 圧縮済みを解凍してから各ヘッダへ飛ばす
	 * @param	bi
	 */
	private function extractCompressedData( bi:BytesInput, fileId:Int ):Void
	{
		var r = new format.gz.Reader( bi );
		var ungz = r.read( );
		var data = new BytesInput( ungz.data );
		data.bigEndian = false;
		this.extractData( data, fileId );
	}

	/**
	 * データ抽出
	 * @param	bi
	 * @param	fileID
	 */
	private function extractData( bi:BytesInput, fileId:Int ):Void
	{
		var saveBinary = #if debug true #else false #end;

		// PSP もっとNUGA-CEL!用
		if ( bi.readInt32( ) != 0x00000020 ) {
			bi.position = 0;
		}

		var header = bi.readString( 4 );
		trace( 'fileId: $fileId, header: $header' );
		switch( header ) {
			case "ART2":
				this.extractART2( bi );
			case "CO2R":
				this.extractCO2R( bi );
			case "CLC2":
				this.extractCLC2( bi );
			case "CL3L":
				this.extractCL3L( bi );
			// case "SDL ":
			//	this.extractSDL( bi );
			default:
				trace( 'shiranai header nanode sonomama hozon suru' );
				saveBinary = true;
		}
		if( saveBinary ) {
			var fo = File.write( this.path + "/UNKNOWN/" + fileId + ".bin" );
			bi.position = 0;
			fo.write( bi.readAll( ) );
			fo.close( );
		}
	}

	/**
	 * ART2を抽出する（PNGとして）
	 * @param	bi
	 */
	private function extractART2( bi:BytesInput )
	{
		var bitDepth = bi.readInt32( );
		if ( bitDepth < 0 ) {
			trace( bitDepth );
			throw 'bitdepth minus toka okashii!!!! chuushi!!!!!!!!';
		}
		var width = bi.readInt32( );
		var height = bi.readInt32( );
		if ( ( width < 0 ) || ( height < 0 ) || ( 1024 <= width ) || ( 1024 <= height )  ) {
			trace( width, height );
			throw 'gazou size ga hentekorin!!!!!! chuushi!!!!!!!!';
		}
		var name = bi.readString( 0x10 ).split( String.fromCharCode( 0x00 ) )[0];
		var len = switch( bitDepth ) {
			case 4: ( width >> 1 ) * height;
			case 8: width * height;
			default:
				throw '$name: bit depth ga hen janai??? size:$bitDepth';
		}

		// 画像を抜く
		var src = new Array<Int>( );
		switch( bitDepth ) {
			case 4:
				for ( i in 0 ... len ) {
					var b = bi.readByte( );
					src.push( b & 0x0f );
					src.push( ( b >> 4 ) & 0x0f );
				}
			case 8:
				for ( i in 0 ... len ) {
					src.push( bi.readByte( ) );
				}
		}

		// パレット取得
		var palette = this.extractPallete( bi, bitDepth );

		// 画像生成
		var bo = new BytesOutput( );
		bo.bigEndian = true;
		for ( t in src ) {
			bo.writeInt32( palette[ t ] );
		}

		// PNGとして保存
		var fo = File.write( this.path + "/ART2/" + this.countART2 + ".png" );
		new Writer(fo).write( Tools.build32ARGB(width, height, bo.getBytes( ) ) );
		fo.close( );
		trace( "ART2 get daze: " + name + " : " + this.countART2 );
		this.countART2 ++;
	}

	/**
	 * CO2Rを抽出（PNGとして）
	 * @param	bi
	 */
	private function extractCO2R( bi:BytesInput )
	{
		// skip 0x10 bytes
		bi.read( 0x10 );

		// カウント
		var count = bi.readInt32( );
		if ( 256 <= count ) {
			throw "meccha gazou ga ooikara hen dato omou!!!!!! dakara chuushi!!!!!!!!!!!!!!!! " + count;
		}
		if ( count < 0 ) {
			throw "gazou maisuu ga minus dakara chuushi!!!!!!!!!!!!!!!! " + count;
		}

		// 名前
		bi.read( 8 );
		var name = bi.readString( 0x10 ).split( String.fromCharCode( 0x00 ) )[0];
		trace( name );

		// ディレクトリ生成
		var dir = this.path + "/CO2R/" + this.countCO2R;
		FileSystem.createDirectory( dir );
		trace( "CO2R get daze: " + this.countCO2R + " / Count:" + count );
		this.countCO2R ++;

		// 画像まで全部無視
		while ( true ) {
			var magic = bi.readString( 4 );
			if ( magic == "LTMR" ) break;
		}
		bi.read( 8 );

		// 画像を全て取る
		for ( imageNumber in 0 ... count ) {
			var len = bi.readInt32( );
			bi.read( 4 );
			var bitDepth = bi.readInt32( );
			var width = bi.readInt32( );
			var height = bi.readInt32( );
			bi.read( 12 );
			var palette = this.extractPallete( bi, 8 );

			// 画像を抜く
			var bo = new BytesOutput( );
			bo.bigEndian = true;
			switch( bitDepth ) {
				case 1:
					for ( i in 0 ... len ) {
						var b = bi.readByte( );
						bo.writeInt32( palette[ b & 0x0f ] );
						bo.writeInt32( palette[ ( b >> 4 ) & 0x0f ] );
					}
				case 2:
					for ( i in 0 ... len ) {
						bo.writeInt32( palette[bi.readByte()] );
					}
			}

			// PNGとして保存
			var fo = File.write( dir + "/" + imageNumber + ".png" );
			new Writer(fo).write( Tools.build32ARGB(width, height, bo.getBytes( ) ) );
			fo.close( );

			trace( '\t1 mai get daze: $imageNumber - $bitDepth bits size($width, $height)' );
		}
	}

	/**
	 * パレット取得
	 * @param	bi
	 * @param	bitDepth
	 */
	private function extractPallete( bi:BytesInput, bitDepth:Int ):Array<Int>
	{
		var size = switch( bitDepth ) {
			case 4: 0x10;
			case 8: 0x100;
			default:
				throw "palette size okashii: " + bitDepth;
		};
		var palette:Array<Int> = [for ( i in 0 ... 0x100 ) 0];

		for ( i in 0 ... size ) {
			var r = bi.readByte( );
			var g = bi.readByte( );
			var b = bi.readByte( );
			var a = Std.int( Math.min( bi.readByte( ) << 1, 0xff ) );
			var argb = (a << 24) | ( r << 16 ) | ( g << 8 ) | b;
			palette[ ( i & 0xe7 ) | ( ( i & 0x10 ) >> 1 ) | ( ( i & 0x08) << 1 ) ] = argb;
		}

		return palette;
	}

	/**
	 * CLC2を抽出する
	 * @param	f
	 */
	private function extractCLC2( bi:BytesInput )
	{
		bi.readInt32( );
		bi.readInt32( );
		bi.readInt32( );

		var size = bi.readInt32( );
		if ( 0x1000000 < size ) {
			throw "ookisugiru node chuushi!!!!";
		}

		// CLC2からテキスト抽出
		var nowCLC2 = this.countCLC2;
		trace( "CLC2 get daze: " + nowCLC2 );
		this.countCLC2 ++;

		// テキストを抽出する
		var allLines = new Array<String>( );
		while ( bi.position < bi.length ) {
			try {
				var header = bi.readInt32( );
				switch( header ) {
					case 0x32545241:	// ART2
						this.extractART2( bi );
					case 0x000000cd:	// メッセージ
						// 常に4
						if ( bi.readInt32( ) != 4 ) continue;
						// 64バイトとばす
						bi.position += 64;
						// テキストのサイズ
						var size = bi.readInt32( );
						if ( 0 < size ) {
							allLines.push( bi.readString( size ).split( String.fromCharCode( 0x00 ) )[0] );
						}
				}
			}catch ( eof:Eof ) {
				break;
			}catch ( d:Dynamic ) {
				trace( d );
			}
		}

		if( 0 < allLines.length ) {
			var fo = File.write( this.path + "/CLC2/" + nowCLC2 + ".txt" );
			fo.writeString( allLines.join( "\r\n" ) );
			fo.close( );
		}
	}

	/**
	 * CL3Lを抽出する
	 * @param	f
	 */
	private function extractCL3L( bi:BytesInput )
	{
		// CLC2からテキスト抽出
		var nowCL3L = this.countCL3L;
		trace( 'CL3L get daze: $nowCL3L' );
		this.countCL3L ++;

		// テキストを抽出する
		var allLines = new Array<String>( );
		while ( bi.position < bi.length ) {
			try {
				var header = bi.readInt32( );
				switch( header ) {
					case 0x32545241:	// ART2
						this.extractART2( bi );
					case 0x000000cd:	// メッセージ
						// 常に4
						if ( bi.readInt32( ) != 4 ) continue;
						// 64バイトとばす
						bi.position += 64;
						// テキストのサイズ
						var size = bi.readInt32( );
						if ( 0 < size ) {
							allLines.push( bi.readString( size ).split( String.fromCharCode( 0x00 ) )[0] );
						}
				}
			}catch ( eof:Eof ) {
				break;
			}catch ( d:Dynamic ) {
				trace( d );
			}
		}

		if( 0 < allLines.length ) {
			var fo = File.write( this.path + "/CL3L/" + nowCL3L + ".txt" );
			fo.writeString( allLines.join( "\r\n" ) );
			fo.close( );
		}
	}
}
